#!/usr/bin/perl -w

use English;
use strict;

use CGI;
use CGI::Carp qw(fatalsToBrowser);

use POSIX qw(strftime);

#
# Configuration
#

my $submaster = ".";   # path to the data
my $dataurl = "data/"; # url to the data directory

#
# End of Configuration
#

my $s = "smadm.cgi";
my $c = "smauth";

my $q = new CGI;
my $cookie = $q->cookie($c);

chdir($submaster)
	or die "can't chdir to $submaster: $!";

if ( defined $q->param("viewsrc") ) {
	print "Content-type: text/plain\n\n";
	open(F, "$s") || die $!;
	while (<F>) { print; }
	close F; exit 0;
}

# password/<username>.txt
#	encrypted password
#
# password/<username>.super
#	if this file exists, the matching user is a superuser
#
# user/<username>.email
#	the e-mail address of that user
#
# open/<year>_<month>_<id>.open
#	empy file which marks a patch as open
#
# data/<year>/<month>/<id>.patch
#	the patch file as originally sent by the user
#
# data/<year>/<month>/<id>.msg
#	messages attached to the patch
#
# data/<year>/<month>/<id>.votes
#	votes by different users
#
# data/<year>/<month>/<id>.owner
#	this patches owner
#
# data/<year>/<month>/<id>.info
#	auto-created additional info
#
# data/<year>/<month>/<id>.done
#	"Applied", "Rejected" or "Discarded"
#	does not exist for open patches
#
# data/<year>/<month>/<id>.deps
#	dependency list for this patch

my $show_patch_counter=0;

sub footer {
	my $info = "&nbsp;";
	$info = "$show_patch_counter Patches" if $show_patch_counter > 1;
	print << "EOT" if not defined $q->param("q");
<table width="100%"><tr><td align=left>
<a href="http://www.rocklinux.org/submaster.html">SubMaster v0.x</a>.
</td><td align=right>$info</td></tr></table>
EOT
}

sub error {
	print "<b>", @_, "</b><hr>\n"; footer(); exit 0;
}

sub getf($) {
	my ($data, $line) = ("", "");
	open(TMP, $_[0]) || error "Can't open $_[0]: $!";
	while ( $line = <TMP> ) { $data .= $line; }
	close TMP; chomp $data; return $data;
}

my ($user, $pass, $super, $action, $id) = ("guest", "guest", 0, "list", "X");

if ( defined $cookie and $cookie=~/^([a-z0-9]+):([a-zA-Z0-9%]+)$/ ) {
	($user, $pass) = ($1, $2);
}

$action = "details" if defined $q->param("i");
$user = $q->param("u") if defined $q->param("u");
$pass = $q->param("p") if defined $q->param("p");
$action = $q->param("a") if defined $q->param("a");

$pass =~ s/([^a-zA-Z0-9%])/sprintf"%%%02X",ord($1)/eg;

if ( $user =~ /^[a-z0-9]+$/ and $pass =~ /^[a-zA-Z0-9%]+$/ ) {
	if ( not defined $cookie or $cookie ne "$user:$pass" ) {
		my $pass_enc = "$pass"; $pass_enc =~ s/%/%25/g;
		$_ = $q->url(-absolute=>1); s,/[^/]*$,/,;
		print "Set-Cookie: $c=$user:$pass_enc; " .
		      "expires=Wednesday, 20-Nov-2013 00:00:00 GMT; path=$_\n";
	}
}

print "Content-type: text/html\n\n";
error "Not allowed character in username!" if $user !~ /^[a-z0-9]+$/;

if ( defined $q->param("newuser") ) {
	error "Username $user is already taken!" if -f "password/$user.txt";
	error "Username guest is reserved!" if $user eq "guest";

	my $p2 = $q->param("p2");
	$p2 =~ s/([^a-zA-Z0-9%])/sprintf"%%%02X",ord($1)/eg;
	error "Passwords are not the same!" if $pass ne $p2;

	system "mkdir -p password";
	system "echo 'Boom!' > password/.htaccess";
	open(F, ">password/$user.txt") || error "Can't open password file";
	print F crypt($pass, chr(65+rand(26)) . chr(65+rand(26))), "\n";
	close F;
}

if ( defined $q->param("i") ) {
	$id = $q->param("i");
	error "Incorrect ID string!"
		if $id !~ /^([0-9]{4})[\/_]?([0-9]{2})[\/_]?([0-9]+)$/;
	$id = "$1/$2/$3";
	error "Unknown ID: $id!" unless -f "data/$id.patch";
}

sub form {
	my $method = $_[0] eq "" ? "get" : "post";
	print "<form action=\"$s\" method=$method>\n";
	print "<input type=hidden name=a value=$_[0]>\n" if $_[0] ne "";
	shift @_; print @_; print "</form>\n";
}

if ( $user eq "guest" ) {
	# do nothing and just let the guest user in
} elsif ( -f "password/$user.txt" ) {
	my $xpass = getf("password/$user.txt");
	goto wrong_password if crypt($pass, $xpass) ne $xpass;
	$super=1 if -f "password/$user.super";
} else {
wrong_password:
	print << "EOT" if not defined $q->param("q");
<title>SubMaster Login</title>
<h1>SubMaster Login</h1>
<form action="$s" method=post>
Username: <input name="u"><br>
Password: <input type="password" name="p">
<input type=submit value="Login">
<p>
Or simply <a href="$s?u=guest&p=guest">log in as guest user</a>.
<p>
You need cookies turned on to use the SubMaster Webfrontend.
</form>
&nbsp;<p><hr><p>
<form action="$s" method=post>
<h3>Create a new User:</h3>
<input type=hidden name="newuser" value="1">
Username: &nbsp;&nbsp; <input name="u"><p>
Password #1: <input type="password" name="p"><br>
Password #2: <input type="password" name="p2">
<input type=submit value="Create">
</form><hr>
EOT
	print "WRONG USERNAME OR PASSWORD!\n" if defined $q->param("q");
	footer();
	exit 0;
}

if ( not defined $q->param("q") ) {
	print "<title>SubMaster Frontend - Logged in as $user</title>\n";
	form "", << "EOT";
<table width="100%"><tr><td>
<h1>SubMaster - Logged in as $user</h1>
<td align=right>
Go to patch #<input type=text name=i>
</td></tr></table>
EOT

	if ( $user eq "guest" ) {
		print << "EOT";
[<a href="$s?a=list">List&nbsp;Open</a>]
[<a href="$s?a=archive">Archive</a>]
[<a href="$s?u=none&p=none">Login / Create User</a>]
<hr>
EOT
	} else {
		print << "EOT";
[<a href="$s?a=list">List&nbsp;Open</a>]
[<a href="$s?a=listmy">List&nbsp;My&nbsp;Open</a>]
[<a href="$s?a=archive">Archive</a>]
[<a href="$s?a=addpatch">Add</a>]
[<a href="$s?a=newpass">Change&nbsp;Pass</a>]
[<a href="$s?a=info">Change&nbsp;Info</a>]
[<a href="$s?u=none&p=none">Logoff</a>]
EOT
		if ( ! -f "user/$user.email" && $action ne "setemail" ) {
			print "<br><b>Warning: <a href=\"$s?a=info\">Your email address is missing in your user info! Please fill it in.</a></b>\n";
		}
		print "<hr>\n";
	}
}

if ( $action eq "newpass" ) {
	error "You are only a guest user!" if $user eq "guest";

	form "setnewpass", << "EOT";
New Password #1: <input type="password" name="p1"><br>
New Password #2: <input type="password" name="p2"><br>
<input type="submit" value="Change Password"><HR>
EOT
}

if ( $action eq "setnewpass" ) {
	error "You are only a guest user!" if $user eq "guest";

	my $p1 = $q->param("p1");
	my $p2 = $q->param("p2");
	$p1 =~ s/([^a-zA-Z0-9%])/sprintf"%%%02X",ord($1)/eg;
	$p2 =~ s/([^a-zA-Z0-9%])/sprintf"%%%02X",ord($1)/eg;
	error "Passwords are not the same!" if $p1 ne $p2;

	open(F, ">password/$user.txt") || error "Can't open password file";
	print F crypt($p1, chr(65+rand(26)) . chr(65+rand(26))), "\n";
	close F;

	print << "EOT";
Password updated. Please click <a href="$s?u=$user&p=$p1">here</a>.
<HR>
EOT
}

if ( $action eq "showuser" ) {
	error "Guest users are not allowed to view user profiles!" if $user eq "guest";

	my $user = $q->param("user");
	error "Hick!" if $user !~ /^[a-z0-9]+$/;

	print "<h3>User-Information for $user:</h3>\n";
	print "<table border cellpadding=3>\n";

	my $data = -f "user/$user.email" ? getf "user/$user.email" : "";
	if ( $data ne "" ) {
		print "<tr><th>Email-Address</th><td>$data</td></tr>\n";
	} else {
		print "<tr><th>Email-Address</th><td>N/A</td></tr>\n";
	}

	$data = -f "user/$user.desc" ? getf "user/$user.desc" : "";
	if ( $data ne "" ) {
		print "<tr><th valign=top>Description</th><td><pre>$data</pre></td></tr>\n";
	} else {
		print "<tr><th>Description</th><td>N/A</td></tr>\n";
	}

	print "</table>\n";
	print "<p>FYI: Users which are not logged in can't see this information.</p><hr>\n";
}

if ( $action eq "setemail" ) {
	error "You are only a guest user!" if $user eq "guest";

	my $email = $q->param("email");
	error "Not allowed character!" if $email !~ /^[a-zA-Z0-9\.\+\@_-]*$/;

	if ( $email eq "" ) {
		unlink "user/$user.email";
	} else {
		system "mkdir -p user";
		system "echo '$email' > user/$user.email";
	}

	$action="info";
}

if ( $action eq "setdesc" ) {
	error "You are only a guest user!" if $user eq "guest";

	my $desc = $q->param("desc");

	if ( $desc !~ /\S/ ) {
		unlink "user/$user.desc";
	} else {
		system "mkdir -p user";
		open(F, ">user/$user.desc") || die $!;
		print F $desc; close F;
	}

	$action="info";
}

if ( $action eq "info" ) {
	error "You are only a guest user!" if $user eq "guest";

	my $email = -f "user/$user.email" ? getf "user/$user.email" : "";
	my $desc = -f "user/$user.desc" ? getf "user/$user.desc" : "";

	form "setemail", << "EOT";
E-Mail:<br><input name="email" size=60 value="$email">
<input type=submit value="Save">
<hr>
EOT
	form "setdesc", << "EOT";
Description:<br><textarea name="desc" cols="40" rows="7">$desc</textarea>
<input type=submit value="Save">
<hr>
EOT
}

if ( $action eq "addpatch" ) {
	error "You are only a guest user!" if $user eq "guest";

	print "<p><form action=\"$s\" method=\"post\" ".
			"enctype=\"multipart/form-data\">\n";
	print "<input type=hidden name=u value=$user>\n";
	print "<input type=hidden name=p value=$pass>\n";
	print "<input type=hidden name=a value=new>\n";
	print "Patch file: <input type=file name=f>\n";
	print "<input type=submit value=Upload>\n";
	print "</form></p>&nbsp;\n";

	print "<p>The patch file should be generated by 'sm patch'\n";
	print "or at least have a compatible format. In general it is\n";
	print "recommended to upload patches using 'sm send' instead\n";
	print "of uploading them here manually.</p>\n<HR>\n";
}

if ( $action eq "new" ) {
	error "You are only a guest user!" if $user eq "guest";

	my $fn = $q->param("f");

	my $patchdata = "";
	while (<$fn>) { s/\r//g; $patchdata .= $_; }

	if ($patchdata !~ /\S/) {
		error "Can't submit an empty patch!";
	}

	$id = strftime "%Y/%m/%d%H%M%S", localtime;
	$id .= sprintf "%05.5d", $$;

	$_ = $id; s/\/[^\/]*$//;
	system("mkdir -p data/$_");

	open F, ">data/$id.patch";
	print F $patchdata;
	close F;

	system "touch data/$id.{msg,info}";
	system "echo '+ $user' > data/$id.votes";
	system "echo $user > data/$id.owner";

	$_ = $id; s/\//_/g;
	system "mkdir -p open";
	system "touch open/$_.open";

	print "Patch $_ added.\n";
	system("./action.sh $id $user new");
	print "<HR>\n" if not defined $q->param("q");
}

sub show_patch($$) {
	my ($id, $show_box) = @_;
	my $state = "Open";
	my $owner = getf("data/$id.owner");

	$show_patch_counter++;
	print "<table width=100%><tr><td valign=top width=80%>\n";
	$_=$id; s/\//_/g;
	if ( ! -f "open/$_.open" ) {
		$state = -f "data/$id.done" ? getf("data/$id.done") : "XXX";
		$state =~ s,Discarded,Discarded/WIP,;
	}
	s/_//g;
	print "<input type=checkbox name=p_$_>\n" if $show_box == 1;
	print "<input type=checkbox name=p_$_ checked>\n" if $show_box == 2;
	print "<b><a href=\"$s?i=$_\">$state Patch</a>: $_</b> [by <a href=\"$s?a=showuser&user=$owner\">$owner</a>]<br/>&nbsp;<br/>\n";

	if ( open(F, "data/$id.patch") ) {
		my $prnewline = 0;
		while (<F>) { last if /\S/; }
		s/([<>])/"&#".ord($1).";"/eg;
		s,\n,<br>\n,g; print;
		while (<F>) {
			last if /^(---|Index:|diff) /;
			if ( ! /\S/ ) { $prnewline=1; next; }
			print "<br>\n" if $prnewline;
			s/([<>])/"&#".ord($1).";"/eg;
			s/\t/        /g; s/ /&nbsp;/g;
			s,\n,<br>\n,g; print;
			$prnewline = 0;
		}
		close F; print "&nbsp;<br/>\n";
	}

	if ( open(F, "data/$id.info") ) {
		my $nl = 0;
		while (<F>) {
			chomp;
			s/\[([0-9]{4}[\/_]?[0-9]{2}[\/_]?[0-9]+)\]/[<a href="$s?i=$1">$1<\/a>]/g;
			s/\(([a-z0-9]+)\)/(<a href="$s?a=showuser&user=$1">$1<\/a>)/g;
			s/((http|https|ftp):\/\/(\S+[a-zA-Z0-9\/]))/<a href=\"$1\">$1<\/a>/g;
			print "$_<br/>\n"; $nl=1;
		}
		close F;
		print "&nbsp;<br/>\n" if $nl;
	}

	if ( open(F, "data/$id.msg") ) {
		while (<F>) {
			chomp; 
			s/\[([0-9]{4}[\/_]?[0-9]{2}[\/_]?[0-9]+)\]/[<a href="$s?i=$1">$1<\/a>]/g;
			s/\(([a-z0-9]+)\)/(<a href="$s?a=showuser&user=$1">$1<\/a>)/g;
			s/((http|https|ftp):\/\/(\S+[a-zA-Z0-9\/]))/<a href=\"$1\">$1<\/a>/g;
			print "$_<br/>\n";
		}
		close F;
	}

	print "</td><td valign=top width=20%>\n";
	if ( not defined $q->param("q") ) {
		my $history_id = $id; $history_id =~ s/\///g;
		my $history_url = "/lurker/cgi-bin/keyword.cgi?doc-url=/lurker&".
		                  "format=html&subject=$history_id&list=";
		print "<nobr>[<a href=\"$dataurl/$id.patch\">View Patch</a>]</nobr> ".
		      "<nobr>[<a href=\"$history_url\">History</a>]</nobr><br/>&nbsp;<br/>\n";
	}

	if ( open(F, "data/$id.votes") ) {
		while (<F>) {
			chomp;
			s/([a-z0-9]+)/<a href="$s?a=showuser&user=$1">$1<\/a>/g;
			print "$_<br/>\n";
		}
		close F;
	}

	if ( open (F, "<data/$id.deps") ) {
		print "<p/>Open Dependencies:<br/>\n";
		while (<F>) {
			chomp; my $plus = -f "data/$_.deps" ? "+" : "";
			s/\//_/g; next unless -f "open/$_.open"; s/_//g;
			print "<a href=\"$s?i=$_\">$_</a>$plus<br/>\n";
		}
		close (F);
	}

	print "</td></tr></table><hr>\n";
}

if ( $action eq "archive" ) {
	my $y = strftime "%Y", localtime;
	my $m = strftime "%m", localtime;
	$y = $1 if defined $q->param("y") and $q->param("y")=~/(\d+)/;
	$m = $1 if defined $q->param("m") and $q->param("m")=~/(\d+)/;

	print "Year:\n";
	while (<data/*>) {
		next unless -d $_ and /^data\/(\d+)$/;
		if ( $1 eq $y ) {
			print "[<b>$1</b>]\n";
		} else {
			print "[<a href=\"$s?a=archive&y=$1\">$1</a>]\n";
		}
	}

	print "<br>\nMonth:\n";
	while (<data/$y/*>) {
		next unless -d $_ and /^data\/\d+\/(\d+)$/;
		if ( $1 eq $m ) {
			print "[<b>$1</b>]\n";
		} else {
			print "[<a href=\"$s?a=archive&y=$y&m=$1\">$1</a>]\n";
		}
	}

	print "<br><hr><p>\n";
	print "<form action=\"$s\" name=\"plist\" method=post>\n";
	while (<data/$y/$m/*.patch>) {
		next unless -f $_ and /^data\/(\d+\/\d+\/\d+)\.patch$/;
		show_patch("$1", 1);
	}
	print "<script language=javascript><!--\n";
	print "function select_all() {\n";
	while (<data/$y/$m/*.patch>) {
		next unless -f $_ and /^data\/(\d+)\/(\d+)\/(\d+)\.patch$/;
		print "\tdocument.plist.p_$1$2$3.checked = !document.plist.p_$1$2$3.checked;\n";
	}
	print "}\n";
	print "function select_my() {\n";
	while (<data/$y/$m/*.votes>) {
		next unless -f $_ and /^data\/(\d+)\/(\d+)\/(\d+)\.votes$/;
		my $line = "\tdocument.plist.p_$1$2$3.checked = !document.plist.p_$1$2$3.checked;\n";
		next unless getf($_) =~ /(^|\n)\+ $user(\n|$)/;
		print $line;
	}
	print "}\n//--></script>\n";
	print "<input type=hidden name=a value=genlist>\n";
	print "<input type=submit value=\"Print selected patch IDs\">\n";
	print "<input type=button onclick=\"select_all()\" value=\"Select/Deselect all\">\n";
	print "<input type=button onclick=\"select_my()\" value=\"Select/Deselect my pro-votes\">\n";
	print "<input type=submit name=\"J\" value=\"Import smap journal\">\n";
	print "</form>\n";
}

if ( $action eq "list" ) {
	print "<form action=\"$s\" name=\"plist\" method=post>\n";
	while (<open/*.open>) {
		next unless m,^open/([0-9]{4})_([0-9]{2})_([0-9]+)\.open$,;
		show_patch("$1/$2/$3", 1);
	}
	print "<script language=javascript><!--\n";
	print "function select_all() {\n";
	while (<open/*.open>) {
		next unless m,^open/([0-9]{4})_([0-9]{2})_([0-9]+)\.open$,;
		print "\tdocument.plist.p_$1$2$3.checked = !document.plist.p_$1$2$3.checked;\n";
	}
	print "}\n";
	print "function select_my() {\n";
	while (<open/*.open>) {
		next unless m,^open/([0-9]{4})_([0-9]{2})_([0-9]+)\.open$,;
		my $line = "\tdocument.plist.p_$1$2$3.checked = !document.plist.p_$1$2$3.checked;\n";
		next unless -f "data/$1/$2/$3.votes" && getf("data/$1/$2/$3.votes") =~ /(^|\n)\+ $user(\n|$)/;
		print $line;
	}
	print "}\n//--></script>\n";
	print "<input type=hidden name=a value=genlist>\n";
	print "<input type=submit value=\"Print selected patch IDs\">\n";
	print "<input type=button onclick=\"select_all()\" value=\"Select/Deselect all\">\n";
	print "<input type=button onclick=\"select_my()\" value=\"Select/Deselect my pro-votes\">\n";
	print "<input type=submit name=\"J\" value=\"Import smap journal\">\n";
	print "</form>\n";
}

if ( $action eq "listmy" ) {
	print "<form action=\"$s\" name=\"plist\" method=post>\n";
	while (<open/*.open>) {
		next unless m,^open/([0-9]{4})_([0-9]{2})_([0-9]+)\.open$,;
		$id = "$1/$2/$3"; $_ = getf("data/$id.owner");
		show_patch($id, 1) if $user eq $_;
	}
	print "<script language=javascript><!--\n";
	print "function select_all() {\n";
	while (<open/*.open>) {
		next unless m,^open/([0-9]{4})_([0-9]{2})_([0-9]+)\.open$,;
		print "\tdocument.plist.p_$1$2$3.checked = !document.plist.p_$1$2$3.checked;\n"
				if getf("data/$1/$2/$3.owner") eq $user;
	}
	print "}\n";
	print "function select_my() {\n";
	while (<open/*.open>) {
		next unless m,^open/([0-9]{4})_([0-9]{2})_([0-9]+)\.open$,;
		my $line = "\tdocument.plist.p_$1$2$3.checked = !document.plist.p_$1$2$3.checked;\n";
		next unless getf("data/$1/$2/$3.owner") eq $user;
		next unless -f "data/$1/$2/$3.votes";
		next unless getf("data/$1/$2/$3.votes") =~ /(^|\n)\+ $user(\n|$)/;
		print $line;
	}
	print "}\n//--></script>\n";
	print "<input type=hidden name=a value=genlist>\n";
	print "<input type=submit value=\"Print selected patch IDs\">\n";
	print "<input type=button onclick=\"select_all()\" value=\"Select/Deselect all\">\n";
	print "<input type=button onclick=\"select_my()\" value=\"Select/Deselect my pro-votes\">\n";
	print "<input type=submit name=\"J\" value=\"Import smap journal\">\n";
	print "</form>\n";
}

if ( $action eq "genlist" ) {
	if ( defined $q->param("J") ) {
		print "<form action=\"$s\" name=\"plist\" method=post>\n";
		print "<input type=hidden name=a value=genlist>\n";
		print "<script><!--\nvar gotfucus_status=1;\nfunction gotfucus() { ";
		print "if ( gotfucus_status ) document.plist.j.value=''; ";
		print "gotfucus_status=0; }\n//--></script>\n";
		print "<textarea cols=60 rows=10 name=\"j\" onFocus=\"gotfucus()\">\n";
		print "[Paste smap journal here.]</textarea>\n";
		print "<br><input type=submit value=\"Print patches from journal\">\n";
		print "</form>\n";
	} else {
		if ( defined $q->param("j") ) {
			my $j = "\n" . $q->param("j") . "\n";
			$j =~ s/#[^\n]*//g;
			while ($j =~ s/\ssmap\s+([0-9]+)\s/ /) {
				$q->param("p_$1", 1);
			}
		}
		print "<b>Seperate smap calls (aka smap journal):</b><p><pre>";
		foreach ($q->param()) {
			next unless s/^p_//;
			print "smap $_\n";
		}
		print "</pre><hr><b>Whitespace-seperated list:</b><p>";
		foreach ($q->param()) {
			next unless s/^p_//;
			print "$_\n";
		}
		print "<hr>";
		print "<form action=\"$s\" name=\"plist\" method=post>\n";
		foreach ($q->param()) {
			next unless /^p_([0-9]{4})([0-9]{2})([0-9]+)$/;
			show_patch("$1/$2/$3", 2);
		}
		print "<input type=hidden name=a value=genlist>\n";
		print "<input type=submit value=\"Print selected patch IDs\">\n";
		print "<script><!--\nvar gotfucus_status=1;\nfunction gotfucus() { ";
		print "if ( gotfucus_status ) document.plist.j.value=''; ";
		print "gotfucus_status=0; }\n//--></script>\n";
		print "<textarea cols=30 rows=1 name=\"j\" onFocus=\"gotfucus()\">\n";
		print "[Paste smap journal here.]</textarea>\n";
		print "</form>\n";
	}
}

if ( $action eq "add_dep" ) {
	my $owner = -f "data/$id.owner" ? getf("data/$id.owner") : "";
	error "This is not your patch!" if $user ne $owner and not $super;

	my $dep = $q->param("dep");
	if ( $dep =~ m,^[0-9]{4}/[0-9]{2}/[0-9]+$, ) {
		system "echo $dep >> data/$id.deps";
	}
	$action = "dependencies";
}

if ( $action eq "del_dep" ) {
	my $owner = -f "data/$id.owner" ? getf("data/$id.owner") : "";
	error "This is not your patch!" if $user ne $owner and not $super;

	my @deps;
	my $dep = $q->param("dep");

	if ( $dep =~ m,^[0-9]{4}/[0-9]{2}/[0-9]+$, ) {
		open (D, "<data/$id.deps");
		while (<D>) {
			chomp;
			if ( $_ ne $dep ) {
				push @deps, $_;
			}
		}
		close (D);
		if ( $#deps >= 0 ) {
			open (D, ">data/$id.deps");
			foreach my $dep (@deps) {
				print D "$dep\n";
			}
			close (D);
		} else {
			unlink("data/$id.deps");
		}
	}
	$action = "dependencies";
}

if ( $action eq "dependencies" and $id ne "X" ) {
	my @deps;
	my $found;
	
	show_patch($id, 0);
	print "<p>This patch depends on:<br />";
	open (D, "<data/$id.deps");
	print "<pre>";
	while (<D>) {
		my $state;
		my $did;
		my $open;
		$state = "Open";
		chomp($did = $_);
		$open = $did;
		$open =~ s/\//_/g;
		if ( ! -f "open/$open.open" ) {
			$state = -f "data/$did.done" ? getf("data/$did.done") : "XXX";
			$state =~ s,Discarded,Discarded/WIP,;
		}
		$did =~ s/\///g;
		print "   <a href=\"$s?i=$did\">$did<\/a> $state\n";
		push @deps, $did;
	}
	close (D);
	print "</pre><hr><p>";
	print "<table cellpadding=2>";
	my $entcounter = 0;
	while (<open/*.open>) {
		next unless m,^open/([0-9]{4})_([0-9]{2})_([0-9]+)\.open$,;
		next if "$1/$2/$3" eq "$id";
		print "<tr><td>&nbsp;</td><td><hr></td></tr>\n" if $entcounter++;
		print "<tr><td valign=\"top\">\n";
		$found=0;
		foreach my $check (@deps){
			if ("$check" == "$1$2$3"){
				$found = 1;
			}
		}
		if ( ! $found ) {
			form "add_dep", "<input type=hidden name=i value=\"$id\"><input type=hidden name=dep value=\"$1/$2/$3\"><input type=\"submit\" value=\"Add\">";
		} else {
			form "del_dep", "<input type=hidden name=i value=\"$id\"><input type=hidden name=dep value=\"$1/$2/$3\"><input type=\"submit\" value=\"Del\">";
		}
		print "</td><td valign=\"top\"><b><a href=\"$s?i=$1/$2/$3\">Open Patch:</a> $1$2$3</b><p/>";
		if ( open(F, "data/$1/$2/$3.patch") ) {
			while (<F>) { last if /\S/; }
			s/([<>])/"&#".ord($1).";"/eg;
			s,\n,<br>\n,g; print;
			while (<F>) {
				last unless /\S/;
				last if /^(---|Index:|diff) /;
				s/([<>])/"&#".ord($1).";"/eg;
				s/\t/        /g; s/ /&nbsp;/g;
				s,\n,<br>\n,g; print;
			}
			close F; print "&nbsp;<br/>\n";
		}
		print "</td>";
		print "</tr>";
	}
	print "</table>";
	print "<hr>";
}

my %print_deps_norec;

sub print_deps($$) {
	my ($id, $space) = @_;
	my @subdeps;

	if ( open (D, "<data/$id.deps") ) {
		print "<p>Dependency tree:<br/><pre>" if not $space;
		$print_deps_norec{$id} = 1;
		push @subdeps, $_ while <D>;
		close (D);
		foreach (@subdeps) {
			my $state;
			my $did;
			my $open;
			$state = "Open";
			chomp($did = $_);
			$open = $did;
			$open =~ s/\//_/g;
			if ( ! -f "open/$open.open" ) {
				$state = -f "data/$did.done" ? getf("data/$did.done") : "XXX";
				$state =~ s,Discarded,Discarded/WIP,;
			}
			$open =~ s/_//g;
			print "   " x $space;
			print "   <a href=\"$s?i=$open\">$open<\/a> $state";
			if ( not defined $print_deps_norec{$did} ) {
				print "\n";
				print_deps($did, $space+1);
			} else {
				print "  (recursion)\n";
			}
		}
		delete $print_deps_norec{$id};
		print "</pre>" if not $space;
		return 1;
	}
	return 0;
}

if ( $action eq "details" and $id ne "X" ) {
	if ( $user eq "guest" ) {
		show_patch($id, 0);
		print "<p><hr><p>\n" if print_deps($id, 0);
		if ((not defined $q->param("q")) && open(F, "data/$id.patch")) {
			my $linenr=0;
			print "<pre>";
			while (<F>) {
				last if ++$linenr == 100;
				s/([<>])/"&#".ord($1).";"/eg;
				print;
			}
			if ( $linenr == 100 ) {
				print "<B>--- Click <a href=\"$dataurl/$id.patch\">here</a> to see the full patch. ---</B>\n";
			} else {
				print "\n<B>[<a href=\"$dataurl/$id.patch\">the plain patch file</a>]</B>\n";
			}
			print "</pre><hr/>\n";
			close F;
		}
		footer();
		exit 0;
	}
	if ( defined $q->param("msg") and open(F, ">>data/$id.msg") ) {
		$_ = $q->param("msg"); s/([<>'"\n\r\t])/"&#".ord($1).";"/eg;
		print F "($user) $_\n" if /\S/; close F;
		system("./action.sh $id $user msg '$_'") if /\S/;
	}
	if ( (defined $q->param("v0") or defined $q->param("v1") or
	      defined $q->param("v2")) and open(F, "data/$id.votes") ) {
		my %votes = ();
		while (<F>) { $votes{$2}=$1 if /^([-+])\s+(\S+)/; }
		$votes{$user} = "-"  if defined $q->param("v0");
		$votes{$user} = "+"  if defined $q->param("v1");
		delete $votes{$user} if defined $q->param("v2");
		close F;
		if ( open(F, ">data/$id.votes") ) {
			foreach ( sort keys %votes ) {
				print F "$votes{$_} $_\n";
			}
			close F;
		}
		system("./action.sh $id $user vote contra") if defined $q->param("v0");
		system("./action.sh $id $user vote pro")    if defined $q->param("v1");
		system("./action.sh $id $user vote delete") if defined $q->param("v2");
	}
	if ( defined $q->param("s0") or defined $q->param("s1") ) {
		error "You are not a super user!" unless $super;
		$_ = $id; s/\//_/g; unlink "open/$_.open";
		$_ = defined $q->param("s0") ? "Rejected" : "Applied";
		system "echo $_ > data/$id.done";
		system("./action.sh $id $user status rejected") if defined $q->param("s0");
		system("./action.sh $id $user status applied")  if defined $q->param("s1");
	}
	if ( defined $q->param("s2") ) {
		error "You are not a super user!" unless $super;
		$_ = $id; s/\//_/g; system "touch open/$_.open";
		unlink "data/$id.done";
		system("./action.sh $id $user status opened");
	}
	if ( defined $q->param("s3") ) {
		if ( not $super ) {
			$_ = getf("data/$id.owner");
			error "You are neighter a super user nor owner of the patch!" unless $_ eq $user;
		}
		$_ = $id; s/\//_/g; unlink "open/$_.open";
		system "echo Discarded > data/$id.done";
		system("./action.sh $id $user status discarded");
	}

	show_patch($id, 0);
	
	my $owner = getf("data/$id.owner");
	if ( $user eq $owner ) {
		$_ = << "EOT";
<br><input type="submit" name="s3" value="Mark this (my) patch as discarded/wip. (E.g. if you want to rediff it.)">
EOT
	} else { $_ = ""; }
	form "details", << "EOT";
<input type=hidden name="i" value="$id">
<p><input type="text" size=60 name="msg">
<input type="submit" value="Add Note">

<p><input type="submit" name="v1" value="Vote for apply (+)">
&nbsp; &nbsp; <input type="submit" name="v0" value="Vote for reject (-)">
&nbsp; &nbsp; <input type="submit" name="v2" value="Clear my vote">$_
<p><hr><p>
EOT
	if ( $user eq $owner or $super ) {
		print_deps($id, 0);
		form "dependencies", << "EOT";
<input type=hidden name="i" value="$id">
<input type=Submit value="Edit dependencies">
EOT
		print "<p><hr><p>\n";
	} else {
		print "<p><hr><p>\n" if print_deps($id, 0);
	}
	if ( $super == 1 ) {
		form "details", << "EOT";
Super-user only:<br>
<input type=hidden name="i" value="$id">
<input type="text" size=60 name="msg"><br>
<input type="submit" name="s1" value="Set Applied">
&nbsp; &nbsp; <input type="submit" name="s0" value="Set Rejected">
&nbsp; &nbsp; <input type="submit" name="s2" value="Set Open">
&nbsp; &nbsp; <input type="submit" name="s3" value="Set Discarded/WIP">
EOT
	}
	print << "EOT";
<hr><p>
<b>Don't use the 'Notes' feature to discuss anything!</b><p>
Discussion should always go to the mailing list, not here. Only use the Notes
feature to (a) make remarks which don't lead to discussion, (b) point to
other patches (i.e. alternative implementations, etc) or (c) to add some
commands for the SubMaster Server Side Scripts.<p>
Patch-ID's in square-brakets (e.g. "[200401101750030313]") are
automatically converted to a link to the specified patch.<hr>
EOT
	if ((not defined $q->param("q")) && open(F, "data/$id.patch")) {
		my $linenr=0;
		print "<pre>";
		while (<F>) {
			last if ++$linenr == 100;
			s/([<>])/"&#".ord($1).";"/eg;
			print;
		}
		if ( $linenr == 100 ) {
			print "<B>--- Click <a href=\"$dataurl/$id.patch\">here</a> to see the full patch. ---</B>\n\n";
		} else {
			print "\n<B>[<a href=\"$dataurl/$id.patch\">the plain patch file</a>]</B>\n\n";
		}
		open(L, "lsdiff -Esn data/$id.patch |") || die $!;
		while (<L>) { print; } close L;
		print "\n<B>[<a href=\"$s?i=$id&a=details&v1=1\">Vote for apply (+)</a>]</B>";
		print "  <B>[<a href=\"$s?i=$id&a=details&v0=1\">Vote for reject (-)</a>]</B>";
		print "</pre><hr/>\n";
		close F;
	}
}

footer();

